/* This file is part of mailfrom filter.
   Copyright (C) 2005, 2006 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
   MA 02110-1301 USA */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <mailutils/mailutils.h>
#include "mailfrom.h"
#define obstack_chunk_alloc malloc
#define obstack_chunk_free free
#include <obstack.h>
#include <string.h>

#include <fprintftime.h>
#ifdef USE_DBM

int
db_list_item(char *dbname, char *email, db_item_printer_t fun)
{	
	DBM_FILE db;
	DBM_DATUM key, contents;
	int rc = 1;
	
	if (mu_dbm_open(dbname, &db, MU_STREAM_READ, 0600, NULL)) {
		mu_error("mu_dbm_open(%s) failed: %s", dbname,
			 mu_strerror(errno));
		return 1;
	}
	memset (&contents, 0, sizeof contents);
	memset(&key, 0, sizeof key);
	MU_DATUM_PTR(key) = email;
	MU_DATUM_SIZE(key) = strlen (email) + 1;
	if (mu_dbm_fetch(&db, key, &contents) == 0) {
		fun(MU_DATUM_PTR(key),
		    MU_DATUM_SIZE(key),
		    MU_DATUM_PTR(contents));
		rc = 0;
	}
	mu_dbm_close(&db);
	return rc;
}

int
db_list(char *dbname, db_item_printer_t fun)
{
	DBM_FILE db;
	DBM_DATUM key, contents;
	int rc = 0;
	
	if (mu_dbm_open(dbname, &db, MU_STREAM_READ, 0600, NULL)) {
		mu_error("mu_dbm_open(%s) failed: %s", dbname,
			 mu_strerror(errno));
		return 1;
	}
	for (key = mu_dbm_firstkey (&db); MU_DATUM_PTR(key);
	     key = mu_dbm_nextkey (&db, key)) {
		memset (&contents, 0, sizeof contents);
		if (mu_dbm_fetch (&db, key, &contents) == 0) {
			fun(MU_DATUM_PTR(key),
			    MU_DATUM_SIZE(key),
			    MU_DATUM_PTR(contents));
		} else {
 			mu_error("Cannot fetch data for %*.*s",
				 MU_DATUM_SIZE(key), MU_DATUM_SIZE(key),
				 MU_DATUM_PTR(key));
			rc = 1;
		}
	}
	mu_dbm_close(&db);
	return 1;
}

int
db_expire(char *dbname, db_expire_t fun)
{
	DBM_FILE db;
	DBM_DATUM key, contents;
	size_t size;
	struct obstack stk;
	char *base, *p;
	size_t i, key_count = 0;
	int rc = 0;
	
	if (mu_dbm_open(dbname, &db, MU_STREAM_RDWR, 0600, NULL)) {
		mu_error("mu_dbm_open(%s) failed: %s", dbname,
			 mu_strerror(errno));
		return 1;
	}

	obstack_init(&stk);
	debug1(1, "Expiring %s: mark phase", dbname);
	for (key = mu_dbm_firstkey (&db); MU_DATUM_PTR(key);
	     key = mu_dbm_nextkey (&db, key)) {
		memset (&contents, 0, sizeof contents);
		if (mu_dbm_fetch (&db, key, &contents) == 0) {
			if (fun(MU_DATUM_PTR(contents))) {
				size = MU_DATUM_SIZE(key);
				obstack_grow(&stk, &size, sizeof size);
				obstack_grow(&stk, MU_DATUM_PTR(key), size);
				key_count++;
			}
		} else {
			mu_error("Cannot fetch data for %*.*s",
				 MU_DATUM_SIZE(key), MU_DATUM_SIZE(key),
				 MU_DATUM_PTR(key));
			rc = !ignore_failed_reads_option;
		}
		mu_dbm_datum_free (&contents);
	}
	size = 0;
	obstack_grow(&stk, &size, sizeof size);

	debug2(1, "Number of keys to expire: %lu, memory used: %u",
	       (unsigned long) key_count,
	       (unsigned) obstack_memory_used(&stk));

	debug1(1, "Expiring %s: sweep phase", dbname);

	base = obstack_finish(&stk);
	for (i = 0, p = base; i < key_count; i++) {
		size = *(size_t*)p;
		p += sizeof(size_t);
		
		debug3(50, "Remove: %*.*s", size, size, p);
		MU_DATUM_PTR(key) = p;
		MU_DATUM_SIZE(key) = size;
		if (mu_dbm_delete(&db, key)) {
			mu_error ("Cannot remove record for %*.*s",
				  size, size, p);
			rc = 1;
		}
		p += size;
	}

	debug1(1, "Finished expiring %s", dbname);
	obstack_free(&stk, NULL);
	mu_dbm_close(&db);
	return rc;
}

int
db_delete(char *dbname, char *id)
{
	DBM_FILE db;
	DBM_DATUM key;
	int rc;
	
	if (mu_dbm_open(dbname, &db, MU_STREAM_RDWR, 0600, NULL)) {
		mu_error("mu_dbm_open(%s) failed: %s", dbname,
			 mu_strerror(errno));
		return 1;
	}
	memset(&key, 0, sizeof key);
	MU_DATUM_PTR(key) = id;
	MU_DATUM_SIZE(key) = strlen (id) + 1;
	if (mu_dbm_delete(&db, key)) {
		mu_error("Cannot remove record for %s", id);
		rc = 1;
	} else
		rc = 0;
	mu_dbm_close(&db);
	return rc;
}	

static char *
make_tmp_name(char *dbname)
{
	char *suf = strrchr(dbname, '.');
	char *p = strrchr(dbname, '/');
	size_t len;
	char *newname;
	char buf[80];
	
	if (p)
		len = p - dbname;
	else if (suf)
		len = suf - dbname;
	else
		len = strlen(dbname);
	snprintf(buf, sizeof buf, "%lu", (unsigned long) getpid());
	newname = emalloc(len + 1 + strlen(buf) + 1);
	memcpy(newname, dbname, len);
	newname[len] = '/';
	p = strcpy(newname + len + 1, buf);
	return newname;
}

int
db_compact(char *dbname, db_expire_t fun)
{
	DBM_DATUM key, contents;
	DBM_FILE odb, ndb;
	int rc = 0;
	char *tmpname;

	debug1(1, "Compacting database `%s'", dbname);
	
	tmpname = make_tmp_name(dbname);

	if (mu_dbm_open(dbname, &odb, MU_STREAM_READ, 0600, NULL)) {
		mu_error("mu_dbm_open(%s) failed: %s", dbname,
			 mu_strerror(errno));
		return 1;
	}

	if (mu_dbm_open(tmpname, &ndb, MU_STREAM_CREAT, 0600, NULL)) {
		mu_error("mu_dbm_open(%s) failed: %s", tmpname,
			 mu_strerror(errno));
		return 1;
	}
	for (key = mu_dbm_firstkey (&odb); MU_DATUM_PTR(key);
	     key = mu_dbm_nextkey (&odb, key)) {
		memset (&contents, 0, sizeof contents);
		if (mu_dbm_fetch (&odb, key, &contents)) {
			rc = !ignore_failed_reads_option;
			mu_error("Cannot fetch data for %*.*s",
				 MU_DATUM_SIZE(key), MU_DATUM_SIZE(key),
				 MU_DATUM_PTR(key));
			continue;
		}
		if (!fun || fun(MU_DATUM_PTR(contents)) == 0) {
			if (((char*)MU_DATUM_PTR(key))[MU_DATUM_SIZE(key) - 1]) {
				/* Old database format. Convert. */
				size_t size;
				char *p;
				DBM_DATUM newkey;

				size = MU_DATUM_SIZE(key);
				p = malloc(size+1);
				if (!MU_DATUM_PTR(newkey)) {
					mu_error("Not enough memory");
					rc = 1;
					break;
				}
				memcpy(p, MU_DATUM_PTR(key), size);
				p[size] = 0;
				
				memset(&newkey, 0, sizeof newkey);
				MU_DATUM_SIZE(newkey) = size+1;
				MU_DATUM_PTR(newkey) = p;
				if (mu_dbm_insert(&ndb, newkey, contents, 1)) {
					mu_error ("Cannot insert datum");
					rc = 1;
				}
				free(p);
			} else if (mu_dbm_insert(&ndb, key, contents, 1)) {
				mu_error ("Cannot insert datum");
				rc = 1;
			}
		}
		mu_dbm_datum_free (&contents);
	}
	mu_dbm_close(&ndb);
	mu_dbm_close(&odb);

	if (rc == 0) {
		if (unlink(dbname)) {
			mu_error("cannot unlink %s: %s",
				 dbname, mu_strerror(errno));
			rc = 1;
		} else if (rename(tmpname, dbname)) {
			mu_error("cannot rename %s to %s: %s",
				 tmpname, dbname, mu_strerror(errno));
			rc = 1;
		}
	}
	
	free(tmpname);
	return rc;
}

size_t
format_time_str(FILE *fp, time_t timestamp)
{
	struct tm tm;

	localtime_r(&timestamp, &tm);
	return fprintftime(fp, time_format_string, &tm, 0, 0);
}

#endif
